# Unix SMB/CIFS implementation.
#
# User model.
#
# Copyright (C) Catalyst.Net Ltd. 2023
#
# Written by Rob van der Linde <rob@catalyst.net.nz>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from ldb import Dn

from samba.dcerpc.security import dom_sid
from samba.dsdb import DS_GUID_USERS_CONTAINER

from .exceptions import NotFound
from .fields import DnField, EnumField, IntegerField, NtTimeField, StringField
from .group import Group
from .org import OrganizationalPerson
from .types import AccountType, UserAccountControl


class User(OrganizationalPerson):
    account_name = StringField("sAMAccountName")
    account_type = EnumField("sAMAccountType", AccountType)
    assigned_policy = DnField("msDS-AssignedAuthNPolicy")
    assigned_silo = DnField("msDS-AssignedAuthNPolicySilo")
    bad_password_time = NtTimeField("badPasswordTime", readonly=True)
    bad_pwd_count = IntegerField("badPwdCount", readonly=True)
    code_page = IntegerField("codePage")
    display_name = StringField("displayName")
    last_logoff = NtTimeField("lastLogoff", readonly=True)
    last_logon = NtTimeField("lastLogon", readonly=True)
    logon_count = IntegerField("logonCount", readonly=True)
    primary_group_id = IntegerField("primaryGroupID")
    pwd_last_set = NtTimeField("pwdLastSet", readonly=True)
    user_account_control = EnumField("userAccountControl", UserAccountControl)
    user_principal_name = StringField("userPrincipalName")

    def __str__(self):
        """Return sAMAccountName rather than cn for User model."""
        return self.account_name

    def get_primary_group(self, samdb) -> Group:
        """Returns the primary Group object for this User."""
        group_sid = f"{samdb.domain_sid}-{self.primary_group_id}"
        return Group.get(samdb, object_sid=group_sid)

    @staticmethod
    def get_base_dn(samdb):
        """Return the base DN for the User model.

        :param samdb: SamDB connection
        :return: Dn to use for new objects
        """
        return samdb.get_wellknown_dn(samdb.get_default_basedn(),
                                      DS_GUID_USERS_CONTAINER)

    @classmethod
    def get_search_dn(cls, samdb):
        """Return Dn used for searching so Computers will also be found.

        :param samdb: SamDB connection
        :return: Dn to use for searching
        """
        return samdb.get_root_basedn()

    @staticmethod
    def get_object_class():
        return "user"

    @classmethod
    def find(cls, samdb, name):
        """Helper function to find a user by Dn, objectSid, or sAMAccountName.

        If the Dn or Sid can't be parsed, use sAMAccountName instead.
        """
        try:
            query = {"dn": Dn(samdb, name)}
        except ValueError:
            try:
                query = {"object_sid": dom_sid(name)}
            except ValueError:
                query = {"account_name": name}

        return cls.get(samdb, **query)

    @classmethod
    def get_sid_for_principal(cls, samdb, principal) -> str:
        """Return object_sid for the provided principal.

        If principal is already an object sid then return without fetching,
        this is different to `User.find` which must fetch the User.
        """
        try:
            return str(dom_sid(principal))
        except ValueError:
            user = cls.find(samdb, principal)
            if user:
                return user.object_sid
            else:
                raise NotFound(f"Principal {principal} not found.")



# Unix SMB/CIFS implementation.
# Copyright (C) Jelmer Vernooij <jelmer@samba.org> 2009-2011
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Tests for samba.netcmd."""

import os
import tempfile

from io import StringIO
from samba.netcmd import Command
from samba.netcmd.testparm import cmd_testparm
from samba.netcmd.main import cmd_sambatool
import samba.tests


class NetCmdTestCase(samba.tests.TestCaseInTempDir):

    def run_netcmd(self, cmd_klass, args, retcode=0):
        cmd = cmd_klass(outf=StringIO(), errf=StringIO())
        cmd.command_name = "apricots"
        try:
            retval = cmd._run(*args)
        except Exception as e:
            cmd.show_command_error(e)
            retval = 1
        self.assertEqual(retcode, retval)
        return cmd.outf.getvalue(), cmd.errf.getvalue()

    def iter_all_subcommands(self):
        todo = list(cmd_sambatool.subcommands.items())
        while todo:
            (path, cmd) = todo.pop()
            yield path, cmd
            subcmds = getattr(cmd, "subcommands", {})
            todo.extend([(path + " " + k, v) for (k, v) in
                         subcmds.items()])


class TestParmTests(NetCmdTestCase):

    def setUp(self):
        super().setUp()

        # Override these global parameters in case their default values are
        # invalid.
        contents = """[global]
    netbios name = test
    lock dir = /
    pid directory = /
[tmp]
    path = /
"""
        self.smbconf = self.create_smbconf(contents)

    def create_smbconf(self, contents):
        smbconf = tempfile.NamedTemporaryFile(mode='w',
                                              dir=self.tempdir,
                                              delete=False)
        self.addCleanup(os.remove, smbconf.name)

        try:
            smbconf.write(contents)
        finally:
            smbconf.close()

        return smbconf

    def test_no_client_ip(self):
        out, err = self.run_netcmd(cmd_testparm, ["--client-name=foo"],
                                   retcode=-1)
        self.assertEqual("", out)
        self.assertEqual(
            "ERROR: Both a DNS name and an IP address are "
            "required for the host access check\n", err)

    def test_section(self):
        # We don't get an opportunity to verify the output, as the parameters
        # are dumped directly to stdout, so just check the return code.
        self.run_netcmd(cmd_testparm,
                        ["--configfile=%s" % self.smbconf.name,
                         "--section-name=tmp"],
                        retcode=None)

    def test_section_globals(self):
        # We can have '[global]' and '[globals]'
        for name in ['global', 'globals']:
            self.run_netcmd(cmd_testparm,
                            [f"--configfile={self.smbconf.name}",
                             f"--section-name={name}"],
                            retcode=None)

    def test_no_such_section(self):
        out, err = self.run_netcmd(cmd_testparm,
                                   ["--configfile=%s" % self.smbconf.name,
                                    "--section-name=foo"],
                                   retcode=-1)
        # Ensure all exceptions are caught.
        self.assertEqual("", out)
        self.assertNotIn("uncaught exception", err)

        out, err = self.run_netcmd(cmd_testparm,
                                   ["--configfile=%s" % self.smbconf.name,
                                    "--section-name=foo",
                                    "--parameter-name=foo"],
                                   retcode=-1)
        # Ensure all exceptions are caught.
        self.assertEqual("", out)
        self.assertNotIn("uncaught exception", err)

    def test_no_such_parameter(self):
        out, err = self.run_netcmd(cmd_testparm,
                                   ["--configfile=%s" % self.smbconf.name,
                                    "--section-name=tmp",
                                    "--parameter-name=foo"],
                                   retcode=-1)
        # Ensure all exceptions are caught.
        self.assertEqual("", out)
        self.assertNotIn("uncaught exception", err)


class CommandTests(NetCmdTestCase):

    def test_description(self):
        class cmd_foo(Command):
            """Mydescription"""
        self.assertEqual("Mydescription", cmd_foo().short_description)

    def test_name(self):
        class cmd_foo(Command):
            pass
        self.assertEqual("foo", cmd_foo().name)

    def test_synopsis_everywhere(self):
        missing = []
        for path, cmd in self.iter_all_subcommands():
            if cmd.synopsis is None:
                missing.append(path)
        if missing:
            self.fail("The following commands do not have a synopsis set: %r" %
                      missing)

    def test_short_description_everywhere(self):
        missing = []
        for path, cmd in self.iter_all_subcommands():
            if cmd.short_description is None:
                missing.append(path)
        if not missing:
            return
        self.fail(
            "The following commands do not have a short description set: %r" %
            missing)
